# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Geode
from geode_gem.widgets.button import GeodeGtkButton
from geode_gem.widgets.common import GeodeGtkCommon
from geode_gem.widgets.misc import GeodeGtkLabel

# GObject
from gi.repository import Gtk, Pango


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class GeodeGtkListBox(GeodeGtkCommon, Gtk.ListBox):

    __setters__ = {
        "set_selection_mode": Gtk.SelectionMode.NONE,
    }

    def __init__(self, *args, **kwargs):
        """ Constructor
        """

        GeodeGtkCommon.__init__(self, Gtk.ListBox, **kwargs)

        self.placeholder = GeodeGtkLabel(
            set_justify=Gtk.Justification.CENTER,
            set_line_wrap=True,
            set_single_line_mode=False,
            set_style=Gtk.STYLE_CLASS_DIM_LABEL,
            set_text=kwargs.get("placeholder", str()))

        # Properties
        self.set_placeholder(self.placeholder)

        # ------------------------------------
        #   Packing
        # ------------------------------------

        self.append_widget(self.placeholder)

        self.placeholder.show()

        for element in args:
            self.add(element)

    def add(self, child):
        """ Append a new child in container

        Parameters
        ----------
        child : Gtk.Widget
            New widget to add into container
        """

        self.append_widget(child)

        super().add(child)

    def do_row_activated(self, row):
        """ Activate inner widget for specified row

        Parameters
        ----------
        row : Gtk.ListBoxRow
            Activated row widget
        """

        if isinstance(row, GeodeGtkListBoxItem):
            row.activate()


class GeodeGtkListBoxItem(GeodeGtkCommon, Gtk.ListBoxRow):

    # Used to specified if Gtk.ListBox.do_activate works with this item
    activatable = False

    def __init__(self, label, *args, **kwargs):
        """ Constructor

        Parameters
        ----------
        """

        GeodeGtkCommon.__init__(self, Gtk.ListBoxRow, **kwargs)

        self.inner_grid = Gtk.Box.new(
            kwargs.get("orientation", Gtk.Orientation.HORIZONTAL),
            kwargs.get("spacing", 12))

        self.label_grid = Gtk.Box.new(Gtk.Orientation.VERTICAL, 2)
        self.label_title = Gtk.Label.new(None)
        self.label_description = Gtk.Label.new(None)

        self.set_text(label, kwargs.get("description", None))

        # ------------------------------------
        #   Properties
        # ------------------------------------

        self.set_activatable(kwargs.get("activatable", True))

        self.inner_grid.set_border_width(kwargs.get("border_width", 6))
        self.inner_grid.set_homogeneous(kwargs.get("homogeneous", False))

        self.label_grid.set_homogeneous(False)

        for widget in (self.label_title, self.label_description):
            widget.set_halign(Gtk.Align.START)
            widget.set_justify(Gtk.Justification.FILL)
            widget.set_line_wrap(True)
            widget.set_line_wrap_mode(Pango.WrapMode.WORD_CHAR)

        self.label_description.set_hexpand(True)
        self.label_description.set_use_markup(True)
        self.label_description.set_no_show_all(True)
        self.label_description.set_valign(Gtk.Align.START)
        self.label_description.get_style_context().add_class(
            Gtk.STYLE_CLASS_DIM_LABEL)

        # ------------------------------------
        #   Packing
        # ------------------------------------

        self.append_widget(self.label_grid)
        self.append_widget(self.label_title)
        self.append_widget(self.label_description)

        self.label_grid.pack_start(self.label_title, True, True, 0)
        self.label_grid.pack_start(self.label_description, True, True, 0)

        self.inner_grid.pack_start(self.label_grid, True, True, 0)

        for element in args:
            self.append_widget(element)
            self.inner_grid.pack_start(
                element,
                getattr(element, "is_fillable", False),
                getattr(element, "is_expandable", False),
                0)

        self.add(self.inner_grid)

    def set_text(self, title, description=None):
        """ Set labels item content

        Parameters
        ----------
        title : str
            Title label string
        description : str, optionnal
            Description label string
        """

        self.label_title.set_label(title)
        self.label_title.set_valign(
            Gtk.Align.END if description is not None else Gtk.Align.CENTER)

        self.label_description.set_visible(description is not None)

        if description is not None:
            self.label_description.set_markup(
                f"<span size='small'>{description}</span>")

    def do_activate(self):
        """ See Gtk.Switch.do_activate()
        """

        if self.get_activatable():
            children = self.inner_grid.get_children()

            if len(children):
                widget = children[-1]

                if isinstance(widget, Gtk.ComboBox):
                    widget.popup()

                elif isinstance(widget, Gtk.FileChooserButton):
                    widget.activate()

                elif isinstance(widget, Gtk.Switch):
                    widget.set_active(not widget.get_active())

                elif issubclass(type(widget), Gtk.Button):
                    widget.clicked()

                elif issubclass(type(widget), Gtk.Entry):
                    widget.grab_focus()

                else:
                    widget.activate()


class GeodeGtkListBoxButton(GeodeGtkListBoxItem):

    activatable = True

    def __init__(self, label, *args, **kwargs):
        """ See geode_gem.ui.widgets.listbox.GeodeGtkListBoxItem
        """

        GeodeGtkListBoxItem.__init__(self, label, *args, **kwargs)

        self.button = GeodeGtkButton(
            kwargs.get("button_label", str()),
            halign=Gtk.Align.END,
            hexpand=False,
            valign=Gtk.Align.CENTER,
            **kwargs)

        # Packing
        self.append_widget(self.button)
        self.inner_grid.pack_end(self.button, False, False, 0)

    def do_activate(self):
        """ See Gtk.Switch.do_activate()
        """

        self.button.clicked()


class GeodeGtkListBoxCheckItem(GeodeGtkListBoxItem):

    activatable = True

    def __init__(self, label, *args, **kwargs):
        """ See geode_gem.ui.widgets.listbox.GeodeGtkListBoxItem
        """

        GeodeGtkListBoxItem.__init__(self, label, *args, **kwargs)

        self.switch = Gtk.Switch.new()
        setattr(self.switch, "identifier", f"{self.identifier}_switch")

        # Properties
        self.switch.set_active(kwargs.get("activate", True))
        self.switch.set_halign(Gtk.Align.END)
        self.switch.set_hexpand(False)
        self.switch.set_valign(Gtk.Align.CENTER)

        # Packing
        self.append_widget(self.switch)

        self.inner_grid.pack_end(self.switch, False, False, 0)

    def do_activate(self):
        """ See Gtk.Switch.do_activate()
        """

        self.switch.activate()



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# GEM
from geode_gem.engine.utils import generate_identifier

# GObject
try:
    from gi.repository import GObject

except ImportError as error:
    from sys import exit

    exit("Cannot found python3-gobject module: %s" % str(error))

# Processus
from subprocess import Popen

# Threading
from threading import Thread


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class ScriptThread(Thread, GObject.GObject):

    __gsignals__ = {
        "script-terminate": (GObject.SignalFlags.RUN_LAST, None, [object]),
    }

    def __init__(self, parent, path, game):
        """ Constructor

        Parameters
        ----------
        parent : gem.interface.Interface
            Main interface to access public variables
        path : str
            Script path
        game : gem.api.Game
            Game object
        """

        Thread.__init__(self)
        GObject.GObject.__init__(self)

        # ------------------------------------
        #   Initialize variables
        # ------------------------------------

        self.parent = parent
        self.logger = parent.logger

        self.path = path
        self.game = game

        self.name = generate_identifier(game.name)

    def run(self):
        """ Launch GameThread instance

        This function start a new processus with Popen and wait until it stop.
        When it finish, ScriptThread emit a signal to main interface.
        """

        try:
            self.proc = Popen([self.path, self.game.name])

            output, error_output = self.proc.communicate()

            self.proc.terminate()

        except OSError as error:
            self.logger.error("Cannot access to script: %s" % str(error))

        except MemoryError as error:
            self.logger.error("A memory error occur: %s" % str(error))

        except KeyboardInterrupt:
            self.logger.info("Terminate by keyboard interrupt")

        except Exception as error:
            self.logger.info("An exception error occur: %s" % str(error))

        # Call game-terminate signal on main window
        self.parent.emit("script-terminate", self)



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# GEM
from geode_gem.ui.widgets.window import CommonWindow

# GObject
from gi.repository import Gtk, Pango

# Translation
from gettext import gettext as _


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class MessageDialog(CommonWindow):

    def __init__(self, parent, title, message, icon, center=True):
        """ Constructor

        Parameters
        ----------
        parent : Gtk.Window
            Parent object
        title : str
            Dialog title
        message : str
            Dialog message
        icon : str
            Default icon name
        center : bool, optional
            If False, use justify text insted of center (Default: True)
        """

        classic_theme = False
        if parent is not None:
            classic_theme = parent.use_classic_theme

        CommonWindow.__init__(self, parent, title, icon, classic_theme)

        # ------------------------------------
        #   Initialize variables
        # ------------------------------------

        self.message = message

        self.center = center

        # ------------------------------------
        #   Prepare interface
        # ------------------------------------

        # Init widgets
        self.__init_widgets()

        # Start interface
        self.__start_interface()

    def __init_widgets(self):
        """ Initialize interface widgets
        """

        self.set_size(480, -1)

        self.set_spacing(6)

        self.set_resizable(True)

        # ------------------------------------
        #   Message
        # ------------------------------------

        text = Gtk.Label()

        # Properties
        text.set_line_wrap(True)
        text.set_use_markup(True)
        text.set_markup(self.message)
        text.set_line_wrap_mode(Pango.WrapMode.WORD)

        if(self.center):
            text.set_justify(Gtk.Justification.CENTER)

        else:
            text.set_justify(Gtk.Justification.FILL)

        # ------------------------------------
        #   Integrate widgets
        # ------------------------------------

        self.pack_start(text, True, True)

    def __start_interface(self):
        """ Load data and start interface
        """

        self.add_button(_("Close"), Gtk.ResponseType.CLOSE, Gtk.Align.END)



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Geode
from geode_gem.widgets.common import GeodeGtkCommon
from geode_gem.widgets.view import CommonView

# GObject
from gi.repository import Gtk


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class GeodeGtkTreeView(CommonView, Gtk.TreeView):

    __setters__ = {
        "set_enable_search": True,
        "set_has_tooltip": False,
        "set_headers_clickable": True,
        "set_headers_visible": True,
        "set_search_column": -1,
        "set_show_expanders": True,
    }

    def __init__(self, model, *args, **kwargs):
        """ Constructor

        Parameters
        ----------
        """

        CommonView.__init__(self, Gtk.TreeView, model, *args, **kwargs)

        # ------------------------------------
        #   Packing
        # ------------------------------------

        for element in args:
            # Only manage TreeViewColumn objects
            if not isinstance(element, Gtk.TreeViewColumn):
                continue

            if self.is_sorterable and element.sort_column_id is not None:
                element.set_sort_column_id(element.sort_column_id)

                if self.sort_func is not None:
                    self.sorted_model.set_sort_func(
                        element.sort_column_id, self.sort_func, element)

            cells = element.get_cells()
            if cells and element.cell_data_func is not None:
                element.set_cell_data_func(cells[0], element.cell_data_func)

            self.append_widget(element)
            self.append_column(element)

        # ------------------------------------
        #   Settings
        # ------------------------------------

        self.set_model(self.inner_model)

    def get_selected_treeiter(self):
        """ See geode_gem.widgets.view.CommonView.get_selected_treeiter
        """

        selection = self.get_selection()

        if selection is not None:
            return selection.get_selected()[1]

        return None

    def select_path_and_scroll(self, treepath, **kwargs):
        """ See geode_gem.widgets.view.CommonView.select_path_and_scroll
        """

        self.set_cursor(treepath, None, kwargs.get("start_editing", False))

        self.scroll_to_cell(treepath,
                            None,
                            kwargs.get("use_align", True),
                            kwargs.get("row_align", 0.5),
                            kwargs.get("col_align", 0.5))

    def set_columns_order(self, *columns):
        """ Set columns order based on column widget keys

        Parameters
        ----------
        columns : list
            Columns widget key list
        """

        for index, key in enumerate(columns):
            if not self.has_widget(key):
                continue

            column_widget = self.get_widget(key)
            self.remove_column(column_widget)
            self.insert_column(column_widget, index)

    def unselect_all(self):
        """ Unselect any selected row from treeview widget
        """

        selection = self.get_selection()

        if selection is not None:
            selection.unselect_all()


class GeodeGtkTreeViewColumn(GeodeGtkCommon, Gtk.TreeViewColumn):

    __setters__ = {
        "set_alignment": 0.5,
        "set_expand": False,
        "set_fixed_width": -1,
        "set_max_width": -1,
        "set_min_width": -1,
        "set_resizable": True,
        "set_reorderable": True,
        "set_sizing": Gtk.TreeViewColumnSizing.GROW_ONLY,
    }

    def __init__(self, title, *args, **kwargs):
        """ Constructor

        Parameters
        ----------
        """

        GeodeGtkCommon.__init__(self, Gtk.TreeViewColumn, **kwargs)

        self.sort_column_id = kwargs.get("sort_column_id", None)

        self.cell_data_func = kwargs.get("cell_data_func", None)

        # ------------------------------------
        #   Properties
        # ------------------------------------

        if title is not None:
            self.set_title(title)

        # ------------------------------------
        #   Packing
        # ------------------------------------

        for element in args:
            self.pack_start(element, element.is_expandable)

            for attribute, value in element.attributes.items():
                self.add_attribute(element, attribute, value)

            self.append_widget(element)



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Filesystem
from pathlib import Path

# GEM
from geode_gem.engine.utils import get_binary_path
from geode_gem.engine.utils import generate_identifier

# System
from shlex import split as shlex_split


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class Emulator(object):

    attributes = {
        "id": str,
        "name": str,
        "default": str,
        "windowed": str,
        "fullscreen": str,
        "icon": Path,
        "binary": Path,
        "configuration": Path,
        "savestates": Path,
        "screenshots": Path
    }

    def __init__(self, **kwargs):
        """ Constructor
        """

        # ----------------------------------------
        #   Initialization
        # ----------------------------------------

        # Initialize variables
        self.__init_keys(**kwargs)

    def __init_keys(self, **kwargs):
        """ Initialize object attributes
        """

        for key, key_type in self.attributes.items():

            if key in kwargs.keys():
                value = kwargs[key]

            elif key_type is Path:
                value = None

            else:
                value = key_type()

            setattr(self, key, value)

            if key_type is Path and type(value) is str:

                path = None
                if len(value) > 0 and not value == "None":
                    path = Path(value).expanduser()

                setattr(self, key, path)

        setattr(self, "id", generate_identifier(self.name))

    def __get_content(self, attribute, game):
        """ Get content list for a specific game

        Parameters
        ----------
        attribute : pathlib.Path
            Emulator path attribute
        game : gem.engine.game.Game
            Game object

        Returns
        -------
        list
            return a files list
        """

        pattern = str(attribute)

        if "<rom_path>" in pattern:
            pattern = pattern.replace("<rom_path>", str(game.path.parent))

        if "<lname>" in pattern:
            pattern = pattern.replace("<lname>", game.path.stem.lower())

        elif "<name>" in pattern:
            pattern = pattern.replace("<name>", game.path.stem)

        if "<key>" in pattern and len(game.key) > 0:
            pattern = pattern.replace("<key>", game.key)

        files = list()

        path = Path(pattern).expanduser().resolve()

        # Check if parent path exists and is a directory
        if path.parent.exists() and path.parent.is_dir():

            for filename in path.parent.glob(path.name):

                # Only retrieve files which exists and are not directories
                if filename.exists() and filename.is_file():
                    files.append(filename)

        return files

    def as_dict(self):
        """ Return object as dictionary structure

        Returns
        -------
        dict
            Data structure
        """

        return {
            "binary": self.binary,
            "configuration": self.configuration,
            "icon": self.icon,
            "save": self.savestates,
            "snaps": self.screenshots,
            "default": self.default,
            "windowed": self.windowed,
            "fullscreen": self.fullscreen
        }

    @property
    def exists(self):
        """ Check if emulator binary exists in user system

        Returns
        -------
        bool
            return True if binary exist, False otherwise
        """

        if len(get_binary_path(self.binary)) > 0:
            return True

        return False

    def get_screenshots(self, game):
        """ Get screenshots list

        Parameters
        ----------
        game : gem.engine.game.Game
            Game object

        Returns
        -------
        list
            Files list

        See Also
        --------
        gem.engine.emulator.Emulator.__get_content()
        """

        return self.__get_content(self.screenshots, game)

    def get_savestates(self, game):
        """ Get savestates list

        Parameters
        ----------
        game : gem.engine.game.Game
            Game object

        Returns
        -------
        list
            Files list

        See Also
        --------
        gem.engine.emulator.Emulator.__get_content()
        """

        return self.__get_content(self.savestates, game)

    def get_command_line(self, game, fullscreen=False):
        """ Generate a launch command

        Parameters
        ----------
        fullscreen : bool, optional
            Use fullscreen parameters (Default: False)

        Returns
        -------
        list or None
            Command launcher parameters list, None otherwise
        """

        # Check emulator binary
        if not self.exists:
            raise FileNotFoundError(
                f"Cannot found emulator binary '{self.binary}'")

        # ----------------------------------------
        #   Retrieve default parameters
        # ----------------------------------------

        arguments = shlex_split(str(self.binary))

        # Retrieve fullscreen mode
        if fullscreen and self.fullscreen is not None:
            arguments.append(f" {self.fullscreen}")

        elif not fullscreen and self.windowed is not None:
            arguments.append(f" {self.windowed}")

        # Retrieve default or specific arguments
        if game.default:
            arguments.append(f" {game.default}")

        elif self.default:
            arguments.append(f" {self.default}")

        # ----------------------------------------
        #   Replace pattern substitutes
        # ----------------------------------------

        command = ' '.join(arguments).strip()

        need_gamefile = True

        keys = {
            "conf_path": self.configuration,
            "rom_name": game.path.stem,
            "rom_file": game.path,
            "rom_path": game.path.parent,
            "key": game.key
        }

        for key, value in keys.items():
            substring = f"<{key}>"

            if value is not None and substring in command:
                command = command.replace(substring, str(value))

                if key in ("rom_path", "rom_name", "rom_file"):
                    need_gamefile = False

        # ----------------------------------------
        #   Generate subprocess compatible command
        # ----------------------------------------

        arguments = shlex_split(command)
        if need_gamefile:
            arguments.append(str(game.path))

        return arguments



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Filesystem
from os import R_OK, W_OK, access

from pathlib import Path

# GEM
from geode_gem.ui.data import Icons
from geode_gem.ui.utils import (on_entry_clear,
                                magic_from_file,
                                on_activate_listboxrow)
from geode_gem.ui.dialog.question import QuestionDialog
from geode_gem.ui.widgets.window import CommonWindow
from geode_gem.ui.widgets.widgets import ListBoxItem

# GObject
from gi import require_version
from gi.repository import Gdk, GLib, Gtk, Pango

# Translation
from gettext import gettext as _

# URL
from urllib.parse import urlparse
from urllib.request import url2pathname


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class EditorDialog(CommonWindow):

    def __init__(self, parent, title, file_path, size, icon, editable=True):
        """ Constructor

        Parameters
        ----------
        parent : Gtk.Window
            Parent object
        title : str
            Dialog title
        file_path : str
            File path
        size : (int, int)
            Dialog size
        icon : gem.data.Icons
            Default icon name
        editable : bool, optional
            If True, allow to modify and save text buffer to file_path
            (Default: True)
        """

        classic_theme = False
        if parent is not None:
            classic_theme = parent.use_classic_theme

        CommonWindow.__init__(self, parent, title, icon, classic_theme)

        # ------------------------------------
        #   Initialize variables
        # ------------------------------------

        if type(editable) is not bool:
            raise TypeError("Wrong type for editable, expected bool")

        self.path = file_path
        self.editable = editable
        self.__width, self.__height = size

        self.founded_iter = list()
        self.current_index = int()
        self.previous_search = str()

        self.modified_buffer = False
        self.refresh_buffer = True

        # ------------------------------------
        #   Targets
        # ------------------------------------

        self.targets = [Gtk.TargetEntry.new("text/uri-list", 0, 1337)]

        # ------------------------------------
        #   Prepare interface
        # ------------------------------------

        # Init widgets
        self.__init_widgets()

        # Init signals
        self.__init_signals()

        # Start interface
        self.__start_interface()

    def __init_widgets(self):
        """ Initialize interface widgets
        """

        self.set_resizable(True)

        self.set_spacing(6)
        self.set_border_width(6)

        # ------------------------------------
        #   Grid
        # ------------------------------------

        self.grid_tools = Gtk.Box()
        self.grid_search = Gtk.Box()

        self.grid_menu_options = Gtk.Box()

        # Properties
        self.grid_tools.set_spacing(12)

        Gtk.StyleContext.add_class(
            self.grid_search.get_style_context(), "linked")

        self.grid_menu_options.set_spacing(6)
        self.grid_menu_options.set_border_width(6)
        self.grid_menu_options.set_orientation(Gtk.Orientation.VERTICAL)

        # ------------------------------------
        #   Path
        # ------------------------------------

        self.entry_path = Gtk.Entry()

        # Properties
        self.entry_path.set_editable(False)
        self.entry_path.set_icon_from_icon_name(
            Gtk.EntryIconPosition.PRIMARY, Icons.Symbolic.TEXT)
        self.entry_path.set_icon_activatable(
            Gtk.EntryIconPosition.PRIMARY, False)

        if not self.editable:
            self.entry_path.set_icon_from_icon_name(
                Gtk.EntryIconPosition.SECONDARY, Icons.Symbolic.REFRESH)

        # ------------------------------------
        #   Menu
        # ------------------------------------

        self.image_menu = Gtk.Image()
        self.button_menu = Gtk.MenuButton()

        self.popover_menu = Gtk.Popover()

        # Properties
        self.image_menu.set_from_icon_name(
            Icons.Symbolic.MENU, Gtk.IconSize.BUTTON)

        self.button_menu.add(self.image_menu)
        self.button_menu.set_use_popover(True)
        self.button_menu.set_popover(self.popover_menu)

        self.popover_menu.add(self.grid_menu_options)
        self.popover_menu.set_modal(True)

        # ------------------------------------
        #   Menu - Options
        # ------------------------------------

        self.frame_menu_options = Gtk.Frame()
        self.listbox_menu_options = Gtk.ListBox()

        self.widget_line = ListBoxItem()
        self.switch_line = Gtk.Switch()

        # Properties
        self.listbox_menu_options.set_activate_on_single_click(True)
        self.listbox_menu_options.set_selection_mode(
            Gtk.SelectionMode.NONE)

        self.widget_line.set_widget(self.switch_line)
        self.widget_line.set_option_label(_("Line break"))

        # ------------------------------------
        #   Menu - Actions
        # ------------------------------------

        self.frame_menu_actions = Gtk.Frame()
        self.listbox_menu_actions = Gtk.ListBox()

        self.widget_import = ListBoxItem()
        self.image_import = Gtk.Image()
        self.button_import = Gtk.Button()

        self.widget_export = ListBoxItem()
        self.image_export = Gtk.Image()
        self.button_export = Gtk.Button()

        # Properties
        self.listbox_menu_actions.set_activate_on_single_click(True)
        self.listbox_menu_actions.set_selection_mode(
            Gtk.SelectionMode.NONE)

        self.button_import.set_image(self.image_import)
        self.button_import.set_relief(Gtk.ReliefStyle.NONE)

        self.image_import.set_from_icon_name(
            Icons.Symbolic.SAVE_AS, Gtk.IconSize.BUTTON)

        self.widget_import.set_widget(self.button_import)
        self.widget_import.set_option_label("%s…" % _("Import"))

        self.button_export.set_image(self.image_export)
        self.button_export.set_relief(Gtk.ReliefStyle.NONE)

        self.image_export.set_from_icon_name(
            Icons.Symbolic.SEND, Gtk.IconSize.BUTTON)

        self.widget_export.set_widget(self.button_export)
        self.widget_export.set_option_label("%s…" % _("Export"))

        # ------------------------------------
        #   Editor
        # ------------------------------------

        self.provider_editor = Gtk.CssProvider.new()

        scroll_editor = Gtk.ScrolledWindow()

        if self.editable:
            try:
                require_version("GtkSource", "4")

                from gi.repository import GtkSource

                self.text_editor = GtkSource.View()
                self.buffer_editor = GtkSource.Buffer()

                self.language_editor = GtkSource.LanguageManager()
                self.style_editor = GtkSource.StyleSchemeManager()

                # Properties
                self.text_editor.set_insert_spaces_instead_of_tabs(True)

                self.buffer_editor.set_language(
                    self.language_editor.guess_language(str(self.path)))

                if self.main_parent is not None:
                    self.text_editor.set_tab_width(
                        self.main_parent.config.getint(
                            "editor", "tab", fallback=4))
                    self.text_editor.set_show_line_numbers(
                        self.main_parent.config.getboolean(
                            "editor", "lines", fallback=False))

                    self.buffer_editor.set_style_scheme(
                        self.style_editor.get_scheme(
                            self.main_parent.config.item(
                                "editor", "colorscheme", "classic")))

            except Exception:
                self.text_editor = Gtk.TextView()
                self.buffer_editor = Gtk.TextBuffer()

        else:
            self.text_editor = Gtk.TextView()
            self.buffer_editor = Gtk.TextBuffer()

        # Properties
        scroll_editor.set_shadow_type(Gtk.ShadowType.OUT)

        self.text_editor.set_editable(self.editable)
        self.text_editor.set_monospace(True)
        self.text_editor.set_top_margin(4)
        self.text_editor.set_left_margin(4)
        self.text_editor.set_right_margin(4)
        self.text_editor.set_bottom_margin(4)
        self.text_editor.set_buffer(self.buffer_editor)
        self.text_editor.drag_dest_set(
            Gtk.DestDefaults.MOTION | Gtk.DestDefaults.DROP, self.targets,
            Gdk.DragAction.COPY)

        self.tag_found = self.buffer_editor.create_tag(
            "found", background="yellow", foreground="black")
        self.tag_current = self.buffer_editor.create_tag(
            "current", background="cyan", foreground="black")

        # ------------------------------------
        #   Search
        # ------------------------------------

        self.entry_search = Gtk.Entry()

        self.image_up = Gtk.Image()
        self.button_up = Gtk.Button()

        self.image_bottom = Gtk.Image()
        self.button_bottom = Gtk.Button()

        # Properties
        self.entry_search.set_placeholder_text(_("Search"))
        self.entry_search.set_icon_from_icon_name(
            Gtk.EntryIconPosition.PRIMARY, Icons.Symbolic.FIND)
        self.entry_search.set_icon_activatable(
            Gtk.EntryIconPosition.PRIMARY, False)
        self.entry_search.set_icon_from_icon_name(
            Gtk.EntryIconPosition.SECONDARY, Icons.Symbolic.CLEAR)

        self.image_up.set_from_icon_name(
            Icons.Symbolic.PREVIOUS, Gtk.IconSize.BUTTON)

        self.button_up.set_image(self.image_up)

        self.image_bottom.set_from_icon_name(
            Icons.Symbolic.NEXT, Gtk.IconSize.BUTTON)

        self.button_bottom.set_image(self.image_bottom)

        # ------------------------------------
        #   Integrate widgets
        # ------------------------------------

        self.grid_tools.pack_start(self.entry_path, True, True, 0)
        self.grid_tools.pack_start(self.grid_search, False, False, 0)
        self.grid_tools.pack_start(self.button_menu, False, False, 0)

        self.pack_start(self.grid_tools, False, False)
        self.pack_start(scroll_editor, True, True)

        scroll_editor.add(self.text_editor)

        self.grid_search.pack_start(self.entry_search, False, True, 0)
        self.grid_search.pack_start(self.button_up, False, True, 0)
        self.grid_search.pack_start(self.button_bottom, False, True, 0)

        self.grid_menu_options.pack_start(
            self.frame_menu_options, False, False, 0)
        self.grid_menu_options.pack_start(
            self.frame_menu_actions, False, False, 0)

        self.frame_menu_options.add(self.listbox_menu_options)

        self.listbox_menu_options.add(self.widget_line)

        self.frame_menu_actions.add(self.listbox_menu_actions)

        if self.editable:
            self.listbox_menu_actions.add(self.widget_import)

        self.listbox_menu_actions.add(self.widget_export)

    def __init_signals(self):
        """ Initialize widgets signals
        """

        self.buffer_editor.connect("changed", self.__on_buffer_modified)

        self.switch_line.connect("state-set", self.__on_break_line)

        self.entry_search.connect("activate", self.__on_entry_update)
        self.entry_search.connect("icon-press", self.__on_entry_clear)

        self.button_bottom.connect("clicked", self.__on_move_search)
        self.button_up.connect("clicked", self.__on_move_search, True)

        self.button_import.connect("clicked", self.__on_import_file)
        self.button_export.connect("clicked", self.__on_export_file)

        self.listbox_menu_options.connect(
            "row-activated", on_activate_listboxrow)
        self.listbox_menu_actions.connect(
            "row-activated", on_activate_listboxrow)

        self.__drop_signal = self.text_editor.connect(
            "drag-data-received", self.__on_dnd_received_data)

        if not self.editable:
            self.entry_path.connect("icon-press", self.__on_refresh_buffer)

            self.window.connect("key-press-event", self.__on_manage_keys)

    def __start_interface(self):
        """ Load data and start interface
        """

        if self.editable:
            self.add_button(_("Cancel"), Gtk.ResponseType.CLOSE)
            self.add_button(_("Save"), Gtk.ResponseType.APPLY, Gtk.Align.END)

        elif self.use_classic_theme:
            self.add_button(_("Close"), Gtk.ResponseType.CLOSE)

        # Parsing font from configuration file
        try:
            font = self.main_parent.config.item(
                "editor", "font", "Sans 12").split()

            # Retrieve informations based on pattern '<family font> <size>'
            font_size = int(font[-1])
            font_family = ' '.join(font[:-1])

            self.provider_editor.load_from_data(bytes(
                "textview { font: %spt '%s' }" % (font_size, font_family),
                "UTF-8"))

            self.text_editor.get_style_context().add_provider(
                self.provider_editor, Gtk.STYLE_PROVIDER_PRIORITY_USER)

        except Exception:
            self.logger.warning("Cannot parse editor font")

        self.entry_path.set_text(str(self.path))

        self.set_size(int(self.__width), int(self.__height))

        self.show_all()
        self.grid_menu_options.show_all()

        self.text_editor.grab_focus()

        self.__on_refresh_buffer()

    def __on_refresh_buffer(self, widget=None, pos=None, event=None):
        """ Load buffer text into editor area

        Parameters
        ----------
        widget : Gtk.Entry, optional
            Entry widget (Default: None)
        pos : Gtk.EntryIconPosition, optional
            Position of the clicked icon (Default: None)
        event : Gdk.EventButton or Gdk.EventKey, optional
            Event which triggered this signal (Default: None)
        """

        if self.refresh_buffer:
            self.refresh_buffer = False

            self.set_subtitle("%s…" % _("Loading"))

            self.window.set_sensitive(False)

            self.buffer_editor.delete(
                self.buffer_editor.get_start_iter(),
                self.buffer_editor.get_end_iter())

            loader = self.__on_load_file()
            self.buffer_thread = GLib.idle_add(loader.__next__)

    def __on_load_file(self):
        """ Load file content into textbuffer
        """

        yield True

        if self.path.exists() and access(self.path, R_OK):
            self.buffer_editor.set_text(self.path.read_text(errors="replace"))

            yield True

        # Remove undo stack from GtkSource.Buffer
        if type(self.buffer_editor) is not Gtk.TextBuffer:
            self.buffer_editor.set_undo_manager(None)

        self.window.set_sensitive(True)

        self.refresh_buffer = True

        self.set_subtitle("")

        yield False

    def __on_manage_keys(self, widget, event):
        """ Manage widgets for specific keymaps

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        event : Gdk.EventButton or Gdk.EventKey
            Event which triggered this signal
        """

        # Refresh buffer
        if event.keyval == Gdk.KEY_F5:
            self.__on_refresh_buffer()

    def __on_break_line(self, widget, status=None):
        """ Set break line mode for textview

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        status : bool or None, optional
            New status for current widget (Default: None)

        Notes
        -----
        Check widget utility in this function
        """

        if status is not None and type(status) is bool:
            widget.set_active(status)

        if status:
            self.text_editor.set_wrap_mode(Gtk.WrapMode.WORD_CHAR)

        else:
            self.text_editor.set_wrap_mode(Gtk.WrapMode.NONE)

    def __init_search(self, text):
        """ Initialize search from search entry

        Parameters
        ----------
        text : str
            Text to match in text buffer
        """

        self.founded_iter = list()

        if len(text) > 0:
            # Remove previous tags from buffer
            self.buffer_editor.remove_all_tags(
                self.buffer_editor.get_start_iter(),
                self.buffer_editor.get_end_iter())

            # Match tags from text in buffer
            self.__on_search_and_mark(
                text, self.buffer_editor.get_start_iter())

    def __on_move_search(self, widget=None, backward=False):
        """ Move between search results

        Parameters
        ----------
        widget : Gtk.Widget, optional
            Object which receive signal (Default: None)
        backward : bool, optional
            If True, use backward search instead of forward (Default: False)
        """

        text = self.entry_search.get_text()

        if self.modified_buffer or not text == self.previous_search:
            # Reset cursor position if different search
            if not text == self.previous_search:
                self.current_index = -1

            self.__init_search(text)

            self.modified_buffer = False

        if len(self.founded_iter) > 0:
            # Avoid to do the same search twice
            self.previous_search = text

            # Avoid to check an index which not exist anymore
            if self.current_index not in range(len(self.founded_iter)):
                self.current_index = -1

            # Remove selector tag from previous match iter
            match = self.founded_iter[self.current_index]

            self.buffer_editor.remove_tag(self.tag_current, match[0], match[1])
            self.buffer_editor.apply_tag(self.tag_found, match[0], match[1])

            if backward:
                self.current_index -= 1

                if self.current_index < 0:
                    self.current_index = len(self.founded_iter) - 1

            else:
                self.current_index += 1

                if self.current_index >= len(self.founded_iter):
                    self.current_index = 0

            # Add selector tag to current match iter
            match = self.founded_iter[self.current_index]

            self.buffer_editor.apply_tag(self.tag_current, match[0], match[1])

            # Scroll editor to current match iter
            self.text_editor.scroll_to_iter(match[0], .25, False, .0, .0)

    def __on_entry_update(self, widget=None):
        """ Search entry value in text buffer

        Parameters
        ----------
        widget : Gtk.Widget, optional
            Object which receive signal
        """

        self.__on_move_search()

    def __on_entry_clear(self, widget, pos, event):
        """ Reset an entry widget when secondary icon is clicked

        Parameters
        ----------
        widget : Gtk.Entry
            Entry widget
        pos : Gtk.EntryIconPosition
            Position of the clicked icon
        event : Gdk.EventButton or Gdk.EventKey
            Event which triggered this signal

        Returns
        -------
        bool
            Function state
        """

        if type(widget) is not Gtk.Entry:
            return False

        if pos == Gtk.EntryIconPosition.SECONDARY \
           and len(widget.get_text()) > 0:
            widget.set_text(str())

            return True

        return False

    def __on_search_and_mark(self, text, start):
        """ Search a text and mark it

        Parameters
        ----------
        text : str
            Text to match in text buffer
        start : Gtk.TextIter
            Start position in text buffer
        """

        match = start.forward_search(text,
                                     Gtk.TextSearchFlags.CASE_INSENSITIVE,
                                     self.buffer_editor.get_end_iter())

        while match is not None:
            self.founded_iter.append(match)

            self.buffer_editor.apply_tag(self.tag_found, match[0], match[1])

            match = match[1].forward_search(
                text, Gtk.TextSearchFlags.CASE_INSENSITIVE,
                self.buffer_editor.get_end_iter())

    def __on_buffer_modified(self, textbuffer):
        """ Check buffer modification

        Parameters
        ----------
        textbuffer : Gtk.TextBuffer
            Modified buffer
        """

        self.modified_buffer = True

    def __on_import_file(self, widget, path=None):
        """ Import a plain text file

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        """

        self.window.set_sensitive(False)

        dialog = ImportDialog(self, self.title, path)

        if dialog.run() == Gtk.ResponseType.APPLY:
            filename = dialog.file_selector.get_filename()

            if filename is not None and len(filename) > 0:
                path = Path(filename).expanduser()

                if path.exists():
                    textbuffer = path.read_text()

                    if dialog.switch_replace.get_active():
                        self.buffer_editor.set_text(textbuffer)

                    else:
                        self.buffer_editor.insert(
                            self.buffer_editor.get_end_iter(), textbuffer)

                    self.__on_entry_update()

        dialog.destroy()

        self.window.set_sensitive(True)

    def __on_export_file(self, widget):
        """ Export buffer content to a plain text file

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        """

        self.window.set_sensitive(False)

        dialog = ExportDialog(self, self.title)

        if dialog.run() == Gtk.ResponseType.APPLY:
            filename = dialog.entry_selector.get_text()

            if len(filename) > 0:
                path = Path(filename).expanduser()

                replace = True

                if path.exists():
                    subdialog = QuestionDialog(
                        self,
                        _("Existing file"),
                        _("Do you want to replace existing file ?"))

                    if subdialog.run() == Gtk.ResponseType.NO:
                        replace = False

                    subdialog.destroy()

                if replace:
                    path.write_text(self.buffer_editor.get_text(
                        self.buffer_editor.get_start_iter(),
                        self.buffer_editor.get_end_iter(), True))

        dialog.destroy()

        self.window.set_sensitive(True)

    def __on_dnd_received_data(self, widget, context, x, y, data, info, time):
        """ Manage drag & drop acquisition

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        context : Gdk.DragContext
            Drag context
        x : int
            X coordinate where the drop happened
        y : int
            Y coordinate where the drop happened
        data : Gtk.SelectionData
            Received data
        info : int
            Info that has been registered with the target in the Gtk.TargetList
        time : int
            Timestamp at which the data was received
        """

        self.text_editor.handler_block(self.__drop_signal)

        # Current acquisition not respect text/uri-list
        if not info == 1337:
            return

        files = data.get_uris()

        if len(files) > 0:
            result = urlparse(files[0])

            if result.scheme == "file":
                path = Path(url2pathname(result.path)).expanduser()

                try:
                    mimetype = magic_from_file(path, mime=True)

                    # Check mimetype format
                    if mimetype is not None and '/' in mimetype:
                        category, *filetype = mimetype.split('/')

                        # Only retrieve text files
                        if category == "text" and path.exists():
                            self.__on_import_file(None, path)

                except Exception as error:
                    self.logger.exception(error)

        self.text_editor.handler_unblock(self.__drop_signal)


class ImportDialog(CommonWindow):

    def __init__(self, parent, title, path=None):
        """ Constructor

        Parameters
        ----------
        parent : Gtk.Window
            Parent object
        title : str
            Window title
        path : str, optional
            Set a default path in import filechooser widget
        """

        classic_theme = False
        if parent is not None:
            classic_theme = parent.use_classic_theme

        CommonWindow.__init__(self,
                              parent,
                              title,
                              Icons.Symbolic.SAVE_AS,
                              classic_theme)

        # ------------------------------------
        #   Variables
        # ------------------------------------

        self.path = path

        self.mimetypes = ["text/plain"]

        # ------------------------------------
        #   Prepare interface
        # ------------------------------------

        # Init widgets
        self.__init_widgets()

        # Init signals
        self.__init_signals()

        # Start interface
        self.__start_interface()

    def __init_widgets(self):
        """ Initialize interface widgets
        """

        self.set_size(520, -1)

        self.set_spacing(6)

        self.set_resizable(True)

        # ------------------------------------
        #   Grid
        # ------------------------------------

        self.grid_switch = Gtk.Grid()

        # Properties
        self.grid_switch.set_column_spacing(12)
        self.grid_switch.set_row_spacing(6)

        # ------------------------------------
        #   Title
        # ------------------------------------

        self.label_title = Gtk.Label()

        # Properties
        self.label_title.set_markup(
            "<span weight='bold' size='large'>%s</span>" % _("Import"))
        self.label_title.set_use_markup(True)
        self.label_title.set_halign(Gtk.Align.CENTER)
        self.label_title.set_ellipsize(Pango.EllipsizeMode.END)

        # ------------------------------------
        #   File selector
        # ------------------------------------

        self.label_selector = Gtk.Label()

        self.filter_selector = Gtk.FileFilter.new()

        self.dialog_selector = Gtk.FileChooserDialog(
            use_header_bar=not self.use_classic_theme)

        self.file_selector = Gtk.FileChooserButton.new_with_dialog(
            self.dialog_selector)

        # Properties
        self.label_selector.set_markup("<b>%s</b>" % _("File"))
        self.label_selector.set_margin_top(12)
        self.label_selector.set_hexpand(True)
        self.label_selector.set_use_markup(True)
        self.label_selector.set_single_line_mode(True)
        self.label_selector.set_halign(Gtk.Align.CENTER)
        self.label_selector.set_ellipsize(Pango.EllipsizeMode.END)

        for pattern in self.mimetypes:
            self.filter_selector.add_mime_type(pattern)

        self.dialog_selector.add_button(
            _("Cancel"), Gtk.ResponseType.CANCEL)
        self.dialog_selector.add_button(
            _("Accept"), Gtk.ResponseType.ACCEPT)
        self.dialog_selector.set_filter(self.filter_selector)
        self.dialog_selector.set_action(Gtk.FileChooserAction.OPEN)
        self.dialog_selector.set_select_multiple(False)
        self.dialog_selector.set_create_folders(False)
        self.dialog_selector.set_local_only(True)

        self.file_selector.set_hexpand(True)

        # ------------------------------------
        #   Optional data
        # ------------------------------------

        self.label_data = Gtk.Label()

        self.frame_options = Gtk.Frame()
        self.scroll_options = Gtk.ScrolledWindow()
        self.listbox_options = Gtk.ListBox()

        self.widget_replace = ListBoxItem()
        self.switch_replace = Gtk.Switch()

        # Properties
        self.label_data.set_markup(
            "<b>%s</b>" % _("Options"))
        self.label_data.set_margin_top(12)
        self.label_data.set_hexpand(True)
        self.label_data.set_use_markup(True)
        self.label_data.set_single_line_mode(True)
        self.label_data.set_halign(Gtk.Align.CENTER)
        self.label_data.set_ellipsize(Pango.EllipsizeMode.END)

        self.listbox_options.set_activate_on_single_click(True)
        self.listbox_options.set_selection_mode(
            Gtk.SelectionMode.NONE)

        self.widget_replace.set_widget(self.switch_replace)
        self.widget_replace.set_option_label(
            _("Replace current buffer"))

        # ------------------------------------
        #   Integrate widgets
        # ------------------------------------

        self.listbox_options.add(self.widget_replace)

        self.scroll_options.add(self.listbox_options)

        self.frame_options.add(self.scroll_options)

        self.pack_start(self.label_title, False, False)
        self.pack_start(self.label_selector, False, False)
        self.pack_start(self.file_selector, False, False)
        self.pack_start(self.label_data, False, False)
        self.pack_start(self.frame_options)

    def __init_signals(self):
        """ Initialize widgets signals
        """

        self.file_selector.connect(
            "file-set", self.__on_file_choose)

        self.listbox_options.connect(
            "row-activated", self.on_activate_listboxrow)

    def __start_interface(self):
        """ Load data and start interface
        """

        self.add_button(_("Cancel"), Gtk.ResponseType.CANCEL)
        self.add_button(_("Apply"), Gtk.ResponseType.APPLY, Gtk.Align.END)

        self.set_default_response(Gtk.ResponseType.APPLY)

        if self.path is not None and self.path.exists():
            self.file_selector.set_filename(str(self.path))

        else:
            self.set_response_sensitive(Gtk.ResponseType.APPLY, False)

    def __on_file_choose(self, *args):
        """ User choose a file with FileChooser
        """

        status = False

        name = self.file_selector.get_filename()

        if name is not None and len(name.strip()) > 0:
            path = Path(name.strip()).expanduser()

            try:
                mimetype = magic_from_file(path, mime=True)

                # Check mimetype format
                if mimetype is not None and '/' in mimetype:
                    category, *filetype = mimetype.split('/')

                    # Only retrieve text files
                    if category == "text" and path.exists():
                        status = True

                    # Unselect this file cause is not a text one
                    else:
                        self.file_selector.unselect_all()

            except Exception as error:
                self.logger.exception(error)

        self.set_response_sensitive(Gtk.ResponseType.APPLY, status)


class ExportDialog(CommonWindow):

    def __init__(self, parent, title):
        """ Constructor

        Parameters
        ----------
        parent : Gtk.Window
            Parent object
        title : str
            Window title
        """

        classic_theme = False
        if parent is not None:
            classic_theme = parent.use_classic_theme

        CommonWindow.__init__(self,
                              parent,
                              title,
                              Icons.Symbolic.SEND,
                              classic_theme)

        # ------------------------------------
        #   Variables
        # ------------------------------------

        self.mimetypes = ["text/plain"]

        # ------------------------------------
        #   Targets
        # ------------------------------------

        self.targets = [Gtk.TargetEntry.new("text/uri-list", 0, 1337)]

        # ------------------------------------
        #   Prepare interface
        # ------------------------------------

        # Init widgets
        self.__init_widgets()

        # Init signals
        self.__init_signals()

        # Start interface
        self.__start_interface()

    def __init_widgets(self):
        """ Initialize interface widgets
        """

        self.set_size(520, -1)

        self.set_spacing(6)

        self.set_resizable(True)

        # ------------------------------------
        #   Grid
        # ------------------------------------

        self.grid_selector = Gtk.Box()

        # Properties
        self.grid_selector.set_orientation(Gtk.Orientation.HORIZONTAL)
        self.grid_selector.set_spacing(6)

        # ------------------------------------
        #   Title
        # ------------------------------------

        self.label_title = Gtk.Label()

        # Properties
        self.label_title.set_markup(
            "<span weight='bold' size='large'>%s</span>" % _("Export"))
        self.label_title.set_use_markup(True)
        self.label_title.set_halign(Gtk.Align.CENTER)
        self.label_title.set_ellipsize(Pango.EllipsizeMode.END)

        # ------------------------------------
        #   File selector
        # ------------------------------------

        self.label_selector = Gtk.Label()

        self.entry_selector = Gtk.Entry()

        self.image_selector = Gtk.Image()
        self.button_selector = Gtk.Button()

        self.filter_selector = Gtk.FileFilter.new()

        # Properties
        self.label_selector.set_markup("<b>%s</b>" % _("File"))
        self.label_selector.set_margin_top(12)
        self.label_selector.set_hexpand(True)
        self.label_selector.set_use_markup(True)
        self.label_selector.set_single_line_mode(True)
        self.label_selector.set_halign(Gtk.Align.CENTER)
        self.label_selector.set_ellipsize(Pango.EllipsizeMode.END)

        self.entry_selector.set_hexpand(True)
        self.entry_selector.set_placeholder_text("%s…" % _("Filepath"))
        self.entry_selector.set_icon_activatable(
            Gtk.EntryIconPosition.PRIMARY, False)
        self.entry_selector.set_icon_from_icon_name(
            Gtk.EntryIconPosition.PRIMARY, Icons.Symbolic.SEND)
        self.entry_selector.set_icon_from_icon_name(
            Gtk.EntryIconPosition.SECONDARY, Icons.Symbolic.CLEAR)
        self.entry_selector.drag_dest_set(
            Gtk.DestDefaults.MOTION | Gtk.DestDefaults.DROP, self.targets,
            Gdk.DragAction.COPY)

        self.image_selector.set_valign(Gtk.Align.CENTER)
        self.image_selector.set_from_icon_name(
            Icons.Symbolic.OPEN, Gtk.IconSize.BUTTON)

        self.button_selector.set_image(self.image_selector)

        for pattern in self.mimetypes:
            self.filter_selector.add_mime_type(pattern)

        # ------------------------------------
        #   Integrate widgets
        # ------------------------------------

        self.grid_selector.pack_start(self.entry_selector, True, True, 0)
        self.grid_selector.pack_start(self.button_selector, False, False, 0)

        self.pack_start(self.label_title, False, False)
        self.pack_start(self.label_selector, False, False)
        self.pack_start(self.grid_selector, False, False)

    def __init_signals(self):
        """ Initialize widgets signals
        """

        self.entry_selector.connect("changed", self.__on_file_choose)
        self.entry_selector.connect("icon-press", on_entry_clear)

        self.__drop_signal = self.entry_selector.connect(
            "drag-data-received", self.__on_dnd_received_data)

        self.button_selector.connect("clicked", self.__on_file_set)

    def __start_interface(self):
        """ Load data and start interface
        """

        self.add_button(_("Cancel"), Gtk.ResponseType.CANCEL)
        self.add_button(_("Apply"), Gtk.ResponseType.APPLY, Gtk.Align.END)

        self.set_default_response(Gtk.ResponseType.APPLY)

        self.set_response_sensitive(Gtk.ResponseType.APPLY, False)

    def __on_file_set(self, *args):
        """ Open a FileChooserDialog to let user choose the export file
        """

        dialog = Gtk.FileChooserDialog(
            title="%s…" % _("Export As"),
            action=Gtk.FileChooserAction.SAVE,
            transient_for=self.window,
            use_header_bar=not self.use_classic_theme)

        dialog.add_buttons(
            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
            Gtk.STOCK_OPEN, Gtk.ResponseType.OK)

        dialog.set_filter(self.filter_selector)
        dialog.set_select_multiple(False)
        dialog.set_create_folders(True)

        name = self.entry_selector.get_text().strip()

        if len(name) == 0:
            path = Path(name).expanduser()

            if not path.exists():
                dialog.set_current_folder(str(Path.home()))

            else:
                dialog.set_filename(str(path))

        else:
            dialog.set_current_folder(str(Path.home()))

        if dialog.run() == Gtk.ResponseType.OK:
            self.entry_selector.set_text(dialog.get_filename())

        dialog.destroy()

    def __on_file_choose(self, *args):
        """ User choose a file with FileChooser
        """

        status = False

        name = self.entry_selector.get_text().strip()

        if len(name) > 0:
            path = Path(name).expanduser()

            status = not path.is_dir() and access(path.parent, W_OK)

        self.set_response_sensitive(Gtk.ResponseType.APPLY, status)

    def __on_dnd_received_data(self, widget, context, x, y, data, info, time):
        """ Manage drag & drop acquisition

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        context : Gdk.DragContext
            Drag context
        x : int
            X coordinate where the drop happened
        y : int
            Y coordinate where the drop happened
        data : Gtk.SelectionData
            Received data
        info : int
            Info that has been registered with the target in the Gtk.TargetList
        time : int
            Timestamp at which the data was received
        """

        self.entry_selector.handler_block(self.__drop_signal)

        # Current acquisition not respect text/uri-list
        if not info == 1337:
            return

        files = data.get_uris()

        if len(files) > 0:
            result = urlparse(files[0])

            if result.scheme == "file":
                path = Path(url2pathname(result.path)).expanduser()

                try:
                    mimetype = magic_from_file(path, mime=True)

                    # Check mimetype format
                    if mimetype is not None and '/' in mimetype:
                        category, *filetype = mimetype.split('/')

                        # Only retrieve text files
                        if category == "text" and path.exists():
                            self.entry_selector.set_text(str(path))

                except Exception as error:
                    self.logger.exception(error)

        self.entry_selector.handler_unblock(self.__drop_signal)



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

from geode_gem.widgets import (button, cellrenderer, combobox, entry, iconview,
                               infobar, listbox, menu, misc, stack, statusbar,
                               toolbar, treeview)


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class GeodeGtk:
    """ Custom widgets for Geode-GEM applications
    """

    Box = misc.GeodeGtkBox
    Button = button.GeodeGtkButton
    ButtonBox = misc.GeodeGtkButtonBox
    CellRendererAccel = cellrenderer.GeodeGtkCellRendererAccel
    CellRendererPixbuf = cellrenderer.GeodeGtkCellRendererPixbuf
    CellRendererText = cellrenderer.GeodeGtkCellRendererText
    CheckMenuItem = menu.GeodeGtkCheckMenuItem
    ComboBox = combobox.GeodeGtkComboBox
    ComboBoxText = combobox.GeodeGtkComboBoxText
    Entry = entry.GeodeGtkEntry
    EntryCompletion = entry.GeodeGtkEntryCompletion
    EntryIcon = entry.GeodeGtkEntryIcon
    FileChooserButton = button.GeodeGtkFileChooserButton
    FontButton = button.GeodeGtkFontButton
    Frame = misc.GeodeGtkFrame
    HeaderBar = misc.GeodeGtkHeaderBar
    IconView = iconview.GeodeGtkIconView
    Image = misc.GeodeGtkImage
    InfoBar = infobar.GeodeGtkInfoBar
    Label = misc.GeodeGtkLabel
    ListBox = listbox.GeodeGtkListBox
    ListBoxButton = listbox.GeodeGtkListBoxButton
    ListBoxCheckItem = listbox.GeodeGtkListBoxCheckItem
    ListBoxItem = listbox.GeodeGtkListBoxItem
    Menu = menu.GeodeGtkMenu
    MenuButton = button.GeodeGtkMenuButton
    MenuItem = menu.GeodeGtkMenuItem
    Overlay = misc.GeodeGtkOverlay
    Popover = misc.GeodeGtkPopover
    RadioMenuItem = menu.GeodeGtkRadioMenuItem
    Revealer = misc.GeodeGtkRevealer
    ScrolledWindow = misc.GeodeGtkScrolledWindow
    SearchEntry = entry.GeodeGtkSearchEntry
    SpinButton = button.GeodeGtkSpinButton
    Spinner = misc.GeodeGtkSpinner
    StackOption = stack.GeodeGtkStackOption
    StackSection = stack.GeodeGtkStackSection
    StackSidebar = stack.GeodeGtkStackSidebar
    StackView = stack.GeodeGtkStackView
    Statusbar = statusbar.GeodeGtkStatusbar
    Switch = misc.GeodeGtkSwitch
    ToggleButton = button.GeodeGtkToggleButton
    Toolbar = toolbar.GeodeGtkToolbar
    ToolbarBox = toolbar.GeodeGtkToolbarBox
    ToolbarSwitch = toolbar.GeodeGtkToolbarSwitch
    TreeView = treeview.GeodeGtkTreeView
    TreeViewColumn = treeview.GeodeGtkTreeViewColumn



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from setuptools import setup
from setuptools import find_packages

setup(
    name='Geode-GEM',
    version='0.11.2',
    author='PacMiam',
    author_email='pacmiam@tuxfamily.org',
    description='Geode-GEM is an interface to manage emulators and games',
    long_description='GEM (Graphical Emulators Manager) is a GTK+ Graphical '
                     'User Interface (GUI) for GNU/Linux which allows you to '
                     'easily manage your emulators. This software aims to '
                     'stay the simplest.',
    keywords='gtk+ emulators games',
    url='https://gem.tuxfamily.org',
    project_urls={
        'Archives': 'https://download.tuxfamily.org/gem/releases',
        'Source': 'https://framagit.org/geode/gem',
        'Tracker': 'https://framagit.org/geode/gem/issues',
    },
    packages=find_packages(exclude=['tools', 'test']),
    include_package_data=True,
    python_requires='~= 3.6',
    install_requires=[
        'PyGobject ~= 3.32',
        'pyxdg ~= 0.26',
    ],
    extras_require={
        'dev': [
            'pytest',
            'flake8',
        ],
    },
    entry_points={
        'console_scripts': [
            'gem-ui = geode_gem.__main__:main',
            'geode-gem = geode_gem.__main__:main',
        ],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: X11 Applications :: GTK',
        'Intended Audience :: End Users/Desktop',
        ('License :: OSI Approved :: '
         'GNU General Public License v3 or later (GPLv3+)'),
        'Natural Language :: French',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Games/Entertainment',
        'Topic :: Utilities',
    ],
)



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Geode
from geode_gem.widgets.common import GeodeGtkCommon

# GObject
from gi.repository import Gtk


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class CommonView(GeodeGtkCommon):

    def __init__(self, subclass, model, *args, **kwargs):
        """ Constructor

        Parameters
        ----------
        model : Gtk.TreeStore
            View model
        """

        GeodeGtkCommon.__init__(self, subclass, **kwargs)

        if not issubclass(type(model), Gtk.TreeModel):
            raise TypeError("Model needs to be a subclass of Gtk.TreeModel")

        self.is_sorterable = kwargs.get("sorterable", False)
        self.is_filterable = kwargs.get("filterable", False)

        self.sorting_column = kwargs.get("sorting_column", None)
        self.sorting_order = kwargs.get(
            "sorting_order", Gtk.SortType.ASCENDING)

        self.sort_func = kwargs.get("sort_func", None)
        self.visible_func = kwargs.get("visible_func", None)

        self.list_model = model
        self.filtered_model = model.filter_new()
        self.sorted_model = Gtk.TreeModelSort(
            model=self.filtered_model if self.is_filterable else model)

        # ------------------------------------
        #   Models
        # ------------------------------------

        self.inner_model = self.list_model

        if self.is_sorterable:
            self.inner_model = self.sorted_model

        elif self.is_filterable:
            self.inner_model = self.filtered_model

        # ------------------------------------
        #   Settings
        # ------------------------------------

        if self.is_sorterable and self.sorting_column is not None:
            self.sorted_model.set_sort_column_id(
                self.sorting_column, self.sorting_order)

        elif self.is_filterable and self.visible_func is not None:
            self.filtered_model.set_visible_func(self.visible_func)

    def append(self, data=None):
        """ Append a new item in main model

        Parameters
        ----------
        data : list, optional
            Data object to append

        Returns
        -------
        Gtk.TreeIter
            Appended row
        """

        return self.list_model.append(row=data)

    def remove(self, treeiter):
        """ Remove a specific iter from main model

        Parameters
        ----------
        treeiter : Gtk.TreeIter
            Iter object to remove
        """

        self.list_model.remove(treeiter)

    def clear(self):
        """ Clear the main model
        """

        self.list_model.clear()

    def convert_child_iter_to_iter(self, treeiter):
        """ Convert a child iter to main model iter

        Parameters
        ----------
        treeiter : Gtk.TreeIter
            Iter object to convert

        Returns
        -------
        Gtk.TreeIter
            Converted iter, None otherwise
        """

        if self.is_filterable:
            status, treeiter = \
                self.filtered_model.convert_child_iter_to_iter(treeiter)

            if not status:
                return None

        if self.is_sorterable:
            status, treeiter = \
                self.sorted_model.convert_child_iter_to_iter(treeiter)

            if not status:
                return None

        return treeiter

    def get_path_from_treeiter(self, treeiter):
        """ Retrieve path from a specific iter from main model

        Parameters
        ----------
        treeiter : Gtk.TreeIter
            Iter object to retrieve

        Returns
        -------
        Gtk.TreePath
            Retrieved iter path, None otherwise
        """

        treeiter = self.convert_child_iter_to_iter(treeiter)

        if treeiter is not None:
            return self.inner_model.get_path(treeiter)

        return None

    def get_selected_treeiter(self):
        """ Retrieve current selected item from icon view

        Returns
        -------
        Gtk.TreeIter
            Selected item, None otherwise
        """

        raise NotImplementedError()

    def get_value(self, treeiter, column):
        """ Retrieve a specific column value from specified treeiter

        Parameters
        ----------
        treeiter : Gtk.TreeIter
            Entry iter from main model
        column : int
            Column index to update

        Returns
        -------
        object
            Column value
        """

        treeiter = self.convert_child_iter_to_iter(treeiter)

        if treeiter is not None:
            return self.inner_model.get_value(treeiter, column)

        return None

    def set_value(self, treeiter, column, value):
        """ Set new value for a specific entry from main model

        Parameters
        ----------
        treeiter : Gtk.TreeIter
            Entry iter from main model
        column : int
            Column index to update
        value : object
            New value to set
        """

        if treeiter is not None and column is not None:
            self.list_model.set_value(treeiter, column, value)

    def select_path_and_scroll(self, treepath, **kwargs):
        """ Select a specific path from view and scroll to the related item

        Parameters
        ----------
        treepath : Gtk.TreePath
            Path value from stored model
        """

        raise NotImplementedError()

    def refilter(self):
        """ Refilter the filtered model if available

        Raises
        ------
        ValueError
            When the filtered model is not available
        """

        if not self.is_filterable:
            raise ValueError(f"{type(self)} is not a filterable widget")

        self.filtered_model.refilter()



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Geode
from geode_gem.widgets.common import GeodeGtkCommon

# GObject
from gi.repository import Gtk, Pango


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class CommonBox(GeodeGtkCommon):

    def __init__(self, subclass, *args, **kwargs):
        """ Constructor

        Parameters
        ----------
        """

        GeodeGtkCommon.__init__(self, subclass, **kwargs)

        # Properties
        if kwargs.get("merge", False):
            Gtk.StyleContext.add_class(self.get_style_context(), "linked")

        # Packing
        method = "pack_start"
        for element in args:
            if element is None:
                method = "pack_end"
                continue

            if not hasattr(self, method):
                method = "add"

            args = [element]
            if not issubclass(type(self), Gtk.HeaderBar) \
               and method.startswith("pack_"):
                args.extend([element.is_expandable, element.is_fillable, 0])

            getattr(self, method)(*args)
            self.append_widget(element)

            element.show_all()


class GeodeGtkBox(CommonBox, Gtk.Box):

    __setters__ = {
        "set_border_width": 0,
        "set_orientation": Gtk.Orientation.HORIZONTAL,
        "set_spacing": 0,
    }

    def __init__(self, *args, **kwargs):
        """ Constructor
        """

        CommonBox.__init__(self, Gtk.Box, *args, **kwargs)


class GeodeGtkButtonBox(CommonBox, Gtk.ButtonBox):

    def __init__(self, *args, **kwargs):
        """ Constructor
        """

        CommonBox.__init__(self, Gtk.ButtonBox, *args, **kwargs)


class GeodeGtkFrame(CommonBox, Gtk.Frame):

    def __init__(self, *args, **kwargs):
        """ Constructor
        """

        CommonBox.__init__(self, Gtk.Frame, *args, **kwargs)


class GeodeGtkHeaderBar(CommonBox, Gtk.HeaderBar):

    def __init__(self, *args, **kwargs):
        """ Constructor
        """

        CommonBox.__init__(self, Gtk.HeaderBar, *args, **kwargs)


class GeodeGtkImage(GeodeGtkCommon, Gtk.Image):

    def __init__(self, *args, **kwargs):
        """ Constructor

        Parameters
        ----------
        """

        GeodeGtkCommon.__init__(self, Gtk.Image, **kwargs)


class GeodeGtkLabel(GeodeGtkCommon, Gtk.Label):

    __setters__ = {
        "set_use_markup": True,
        "set_use_underline": True,
        "set_ellipsize": Pango.EllipsizeMode.NONE,
    }

    def __init__(self, *args, **kwargs):
        """ Constructor
        """

        GeodeGtkCommon.__init__(self, Gtk.Label, **kwargs)


class GeodeGtkOverlay(GeodeGtkCommon, Gtk.Overlay):

    def __init__(self, *args, **kwargs):
        """ Constructor
        """

        GeodeGtkCommon.__init__(self, Gtk.Overlay, **kwargs)

        for element in args:
            self.append_widget(element)
            self.add(element)
            element.show_all()

        if "overlay" in kwargs:
            widget = kwargs.get("overlay")

            self.append_widget(widget)
            self.add_overlay(widget)
            widget.show_all()


class GeodeGtkPopover(CommonBox, Gtk.Popover):

    def __init__(self, *args, **kwargs):
        """ Constructor
        """

        CommonBox.__init__(self, Gtk.Popover, *args, **kwargs)


class GeodeGtkRevealer(CommonBox, Gtk.Revealer):

    def __init__(self, *args, **kwargs):
        """ Constructor
        """

        CommonBox.__init__(self, Gtk.Revealer, *args, **kwargs)


class GeodeGtkScrolledWindow(GeodeGtkCommon, Gtk.ScrolledWindow):

    def __init__(self, *args, **kwargs):
        """ Constructor
        """

        CommonBox.__init__(self, Gtk.ScrolledWindow, *args, **kwargs)


class GeodeGtkSpinner(GeodeGtkCommon, Gtk.Spinner):

    def __init__(self, *args, **kwargs):
        """ Constructor
        """

        GeodeGtkCommon.__init__(self, Gtk.Spinner, **kwargs)


class GeodeGtkSwitch(GeodeGtkCommon, Gtk.Switch):

    __setters__ = {
        "set_active": True,
        "set_hexpand": False,
        "set_valign": Gtk.Align.CENTER,
    }

    def __init__(self, *args, **kwargs):
        """ Constructor
        """

        GeodeGtkCommon.__init__(self, Gtk.Switch, **kwargs)



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Geode
from geode_gem.widgets.common import GeodeGtkCommon

# GObject
from gi.repository import GdkPixbuf, Gio, Gtk


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class CommonEntry(GeodeGtkCommon):

    def __init__(self, subclass, *args, **kwargs):
        """ Constructor

        Parameters
        ----------
        subclass : Gtk.Entry
            Subclass widget type
        """

        GeodeGtkCommon.__init__(self, subclass, **kwargs)

        # Packing
        for element in args:
            self.append_widget(element)

    def set_completion_data(self, identifier, *args):
        """ Set entry completion data

        Parameters
        ----------
        identifier : str
            String to identify this object in internal container
        """

        if not self.get_completion() or not self.has_widget(identifier):
            return

        widget = self.get_widget(identifier)

        widget.list_model.clear()
        for element in args:
            widget.list_model.append([element])

    def append_widget(self, widget, *args, **kwargs):
        """ See geode_gem.widgets.common.GeodeGtkCommon.append_widget

        Allow to manage completion and entry icons
        """

        super().append_widget(widget, *args, **kwargs)

        if isinstance(widget, GeodeGtkEntryCompletion):
            self.set_completion(widget)

        elif isinstance(widget, GeodeGtkEntryIcon):
            method = "set_icon_from_icon_name"

            if isinstance(widget.widget, Gio.Icon):
                method = "set_icon_from_gicon"

            elif isinstance(widget.widget, GdkPixbuf.Pixbuf):
                method = "set_icon_from_pixbuf"

            getattr(self, method)(widget.position, widget.widget)
            self.set_icon_tooltip_markup(widget.position, widget.tooltip_text)
            self.set_icon_activatable(widget.position, widget.is_activatable)


class GeodeGtkEntry(CommonEntry, Gtk.Entry):

    def __init__(self, *args, **kwargs):
        """ Constructor
        """

        CommonEntry.__init__(self, Gtk.Entry, *args, **kwargs)


class GeodeGtkEntryCompletion(GeodeGtkCommon, Gtk.EntryCompletion):

    __setters__ = {
        "set_match_func": None,
        "set_popup_completion": True,
        "set_popup_single_match": True,
    }

    def __init__(self, model, *args, **kwargs):
        """ Constructor

        Parameters
        ----------
        model : Gtk.TreeStore
            Completion model
        """

        GeodeGtkCommon.__init__(self, Gtk.EntryCompletion, **kwargs)

        self.list_model = model

        # Settings
        self.set_model(model)


class GeodeGtkEntryIcon():

    def __init__(self, position, widget, **kwargs):
        """ Constructor

        Parameters
        ----------
        position : Gtk.EntryIconPosition
            Icon position in Gtk.Entry widget
        widget : Gio.Icon or str or GdkPixbuf.Pixbuf
            Icon widget
        """

        self.position = position
        self.widget = widget

        # Properties
        self.tooltip_text = kwargs.get("tooltip", str())
        self.is_activatable = kwargs.get("is_activatable", False)


class GeodeGtkSearchEntry(CommonEntry, Gtk.SearchEntry):

    def __init__(self, *args, **kwargs):
        """ Constructor
        """

        CommonEntry.__init__(self, Gtk.SearchEntry, *args, **kwargs)



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Geode
from geode_gem.widgets.common import GeodeGtkCommon
from geode_gem.widgets.misc import GeodeGtkLabel

# GObject
from gi.repository import Gtk


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class GeodeGtkInfoBar(GeodeGtkCommon, Gtk.InfoBar):

    __setters__ = {
        "set_show_close_button": False,
    }

    def __init__(self, *args, **kwargs):
        """ Constructor
        """

        GeodeGtkCommon.__init__(self, Gtk.InfoBar, **kwargs)

        self.inner_grid = self.get_content_area()

        self.label = GeodeGtkLabel(set_use_markup=True)

        # Packing
        self.append_widget(self.label)
        self.inner_grid.pack_start(self.label, True, True, 4)

    def set_message(self, message_type, message_text):
        """ Set a new message with a specific type

        Parameters
        ----------
        message_type : Gtk.MessageType
            Infobar message type
        message_text : str
            The new infobar label text

        Raises
        ------
        TypeError
            When specified message_type type is not a Gtk.MessageType
        """

        if not isinstance(message_type, Gtk.MessageType):
            raise TypeError(
                "The message_type parameter must be a Gtk.MessageType")

        self.set_message_type(message_type)
        if message_type and not self.get_visible():
            self.set_visible(True)

        self.label.set_markup(message_text)



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# GEM
from geode_gem.ui.data import Icons
from geode_gem.ui.widgets.window import CommonWindow

# GObject
from gi.repository import Gdk, GdkPixbuf, GLib, Gtk


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class ViewerDialog(CommonWindow):

    def __init__(self, parent, title, size, screenshots_path):
        """ Constructor

        Parameters
        ----------
        parent : Gtk.Window
            Parent object
        title : str
            Dialog title
        size : (int, int)
            Dialog size
        screenshots_path : list
            Screnshots path list
        """

        classic_theme = False
        if parent is not None:
            classic_theme = parent.use_classic_theme

        CommonWindow.__init__(self, parent, title, Icons.IMAGE, classic_theme)

        # ------------------------------------
        #   Initialize variables
        # ------------------------------------

        self.index = 0

        self.zoom_fit = True

        self.zoom_min = 10
        self.zoom_max = 400
        self.zoom_step = 5
        self.zoom_page_step = 10
        self.zoom_actual = 100

        # Allow the picture to autosize (with zoom_fit) when resize dialog
        self.auto_resize = parent.config.getboolean(
            "viewer", "auto_resize", fallback=False)

        self.screenshots = screenshots_path

        self.current_path = None
        self.current_pixbuf = None
        self.current_pixbuf_zoom = None
        self.current_pixbuf_size = tuple()

        self.targets = parent.targets

        self.__width, self.__height = size

        # ------------------------------------
        #   Manage monitor geometry
        # ------------------------------------

        self.__default_size = 800

        # Get default display
        display = Gdk.Display.get_default()

        if display is not None:
            # Retrieve default display monitor
            monitor = display.get_primary_monitor()
            # Retrieve default monitor geometry
            geometry = monitor.get_geometry()

            self.__default_size = min(
                int(geometry.width / 2), int(geometry.height / 2))

        # ------------------------------------
        #   Prepare interface
        # ------------------------------------

        # Init widgets
        self.__init_widgets()

        # Init signals
        self.__init_signals()

        # Start interface
        self.__start_interface()

    def __init_widgets(self):
        """ Initialize interface widgets
        """

        self.set_resizable(True)

        self.set_spacing(0)
        self.set_border_width(0)

        self.grid_tools.set_border_width(6)

        # ------------------------------------
        #   Overlay
        # ------------------------------------

        self.overlay = Gtk.Overlay()

        # ------------------------------------
        #   Image
        # ------------------------------------

        self.scroll_image = Gtk.ScrolledWindow()
        self.view_image = Gtk.Viewport()

        self.image = Gtk.Image()

        # Properties
        self.view_image.drag_source_set(
            Gdk.ModifierType.BUTTON1_MASK, self.targets, Gdk.DragAction.COPY)

        # ------------------------------------
        #   Resize buttons
        # ------------------------------------

        self.grid_resize_buttons = Gtk.Box()

        self.image_fit = Gtk.Image()
        self.button_fit = Gtk.Button()

        self.image_original = Gtk.Image()
        self.button_original = Gtk.Button()

        # Properties
        Gtk.StyleContext.add_class(
            self.grid_resize_buttons.get_style_context(), "linked")
        self.grid_resize_buttons.set_spacing(-1)
        self.grid_resize_buttons.set_orientation(Gtk.Orientation.HORIZONTAL)

        self.image_fit.set_from_icon_name(
            Icons.Symbolic.ZOOM_FIT, Gtk.IconSize.BUTTON)

        self.button_fit.set_image(self.image_fit)

        self.image_original.set_from_icon_name(
            Icons.Symbolic.ZOOM, Gtk.IconSize.BUTTON)

        self.button_original.set_image(self.image_original)

        # ------------------------------------
        #   Move buttons
        # ------------------------------------

        self.grid_move_buttons = Gtk.Box()

        self.image_first = Gtk.Image()
        self.button_first = Gtk.Button()

        self.image_previous = Gtk.Image()
        self.button_previous = Gtk.Button()

        self.image_next = Gtk.Image()
        self.button_next = Gtk.Button()

        self.image_last = Gtk.Image()
        self.button_last = Gtk.Button()

        # Properties
        Gtk.StyleContext.add_class(
            self.grid_move_buttons.get_style_context(), "linked")
        self.grid_move_buttons.set_spacing(-1)
        self.grid_move_buttons.set_orientation(Gtk.Orientation.HORIZONTAL)

        self.image_first.set_from_icon_name(
            Icons.Symbolic.FIRST, Gtk.IconSize.BUTTON)

        self.button_first.set_image(self.image_first)

        self.image_previous.set_from_icon_name(
            Icons.Symbolic.PREVIOUS, Gtk.IconSize.BUTTON)

        self.button_previous.set_image(self.image_previous)

        self.image_next.set_from_icon_name(
            Icons.Symbolic.NEXT, Gtk.IconSize.BUTTON)

        self.button_next.set_image(self.image_next)

        self.image_last.set_from_icon_name(
            Icons.Symbolic.LAST, Gtk.IconSize.BUTTON)

        self.button_last.set_image(self.image_last)

        # ------------------------------------
        #   Zoom buttons
        # ------------------------------------

        self.grid_zoom_buttons = Gtk.Box()

        self.image_zoom_minus = Gtk.Image()
        self.button_zoom_minus = Gtk.Button()

        self.image_zoom_plus = Gtk.Image()
        self.button_zoom_plus = Gtk.Button()

        self.adjustment_zoom = Gtk.Adjustment()

        self.scale_zoom = Gtk.Scale()

        # Properties
        Gtk.StyleContext.add_class(
            self.grid_zoom_buttons.get_style_context(), "linked")
        self.grid_zoom_buttons.set_spacing(-1)
        self.grid_zoom_buttons.set_orientation(Gtk.Orientation.HORIZONTAL)

        self.image_zoom_minus.set_from_icon_name(
            Icons.Symbolic.ZOOM_OUT, Gtk.IconSize.BUTTON)

        self.button_zoom_minus.set_image(self.image_zoom_minus)

        self.image_zoom_plus.set_from_icon_name(
            Icons.Symbolic.ZOOM_IN, Gtk.IconSize.BUTTON)

        self.button_zoom_plus.set_image(self.image_zoom_plus)

        self.adjustment_zoom.set_lower(self.zoom_min)
        self.adjustment_zoom.set_upper(self.zoom_max)
        self.adjustment_zoom.set_step_increment(self.zoom_step)
        self.adjustment_zoom.set_page_increment(self.zoom_page_step)

        self.scale_zoom.set_draw_value(False)
        self.scale_zoom.set_size_request(150, -1)
        self.scale_zoom.set_adjustment(self.adjustment_zoom)
        self.scale_zoom.add_mark(self.zoom_min, Gtk.PositionType.BOTTOM, None)
        self.scale_zoom.add_mark(200, Gtk.PositionType.BOTTOM, None)
        self.scale_zoom.add_mark(self.zoom_max, Gtk.PositionType.BOTTOM, None)

        # ------------------------------------
        #   Overlay move buttons
        # ------------------------------------

        self.image_previous = Gtk.Image()
        self.button_overlay_previous = Gtk.Button()

        self.image_next = Gtk.Image()
        self.button_overlay_next = Gtk.Button()

        # Properties
        self.image_previous.set_from_icon_name(
            Icons.Symbolic.PREVIOUS, Gtk.IconSize.BUTTON)

        self.button_overlay_previous.get_style_context().add_class("osd")
        self.button_overlay_previous.set_image(self.image_previous)
        self.button_overlay_previous.set_valign(Gtk.Align.CENTER)
        self.button_overlay_previous.set_halign(Gtk.Align.START)
        self.button_overlay_previous.set_no_show_all(True)
        self.button_overlay_previous.set_margin_bottom(6)
        self.button_overlay_previous.set_margin_start(6)
        self.button_overlay_previous.set_margin_end(6)
        self.button_overlay_previous.set_margin_top(6)

        self.image_next.set_from_icon_name(
            Icons.Symbolic.NEXT, Gtk.IconSize.BUTTON)

        self.button_overlay_next.get_style_context().add_class("osd")
        self.button_overlay_next.set_image(self.image_next)
        self.button_overlay_next.set_valign(Gtk.Align.CENTER)
        self.button_overlay_next.set_halign(Gtk.Align.END)
        self.button_overlay_next.set_no_show_all(True)
        self.button_overlay_next.set_margin_bottom(6)
        self.button_overlay_next.set_margin_start(6)
        self.button_overlay_next.set_margin_end(6)
        self.button_overlay_next.set_margin_top(6)

        # ------------------------------------
        #   Integrate widgets
        # ------------------------------------

        self.scroll_image.add(self.view_image)
        self.view_image.add(self.image)

        self.grid_move_buttons.pack_start(
            self.button_first, False, False, 0)
        self.grid_move_buttons.pack_start(
            self.button_previous, False, False, 0)
        self.grid_move_buttons.pack_start(
            self.button_next, False, False, 0)
        self.grid_move_buttons.pack_start(
            self.button_last, False, False, 0)

        # self.grid_zoom_buttons.pack_start(
        #    self.button_zoom_minus, False, False, 0)
        self.grid_zoom_buttons.pack_start(
            self.button_original, False, False, 0)
        self.grid_zoom_buttons.pack_start(
            self.button_fit, False, False, 0)
        # self.grid_zoom_buttons.pack_start(
        #    self.button_zoom_plus, False, False, 0)

        self.overlay.add(self.scroll_image)
        self.overlay.add_overlay(self.button_overlay_previous)
        self.overlay.add_overlay(self.button_overlay_next)

        self.pack_start(self.overlay, True, True, 0)

    def __init_signals(self):
        """ Initialize widgets signals
        """

        if self.auto_resize:
            self.window.connect("size-allocate", self.update_screenshot)

        self.window.connect("key-press-event", self.change_screenshot)

        self.button_first.connect("clicked", self.change_screenshot)
        self.button_previous.connect("clicked", self.change_screenshot)
        self.button_next.connect("clicked", self.change_screenshot)
        self.button_last.connect("clicked", self.change_screenshot)

        self.button_zoom_minus.connect("clicked", self.change_screenshot)
        self.button_original.connect("clicked", self.change_screenshot)
        self.button_fit.connect("clicked", self.change_screenshot)
        self.button_zoom_plus.connect("clicked", self.change_screenshot)

        self.button_overlay_previous.connect("clicked", self.change_screenshot)
        self.button_overlay_next.connect("clicked", self.change_screenshot)

        self.view_image.connect("drag-data-get", self.__on_dnd_send_data)

        self.scale_zoom.connect("change_value", self.update_adjustment)

    def __start_interface(self):
        """ Load data and start interface
        """

        self.add_widget(self.grid_move_buttons)
        self.add_widget(self.grid_zoom_buttons, Gtk.Align.END)
        self.add_widget(self.scale_zoom, Gtk.Align.END)

        self.set_size(int(self.__width), int(self.__height))

        self.show_all()

        self.update_screenshot()
        self.set_widgets_sensitive()

    def __on_dnd_send_data(self, widget, context, data, info, time):
        """ Set screenshot file path uri

        This function send rom file path uri when user drag a game from gem and
        drop it to extern application

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        context : Gdk.DragContext
            Drag context
        data : Gtk.SelectionData
            Received data
        info : int
            Info that has been registered with the target in the Gtk.TargetList
        time : int
            Timestamp at which the data was received
        """

        if self.current_path is not None and self.current_path.exists():
            data.set_uris(["file://%s" % str(self.current_path)])

    def change_screenshot(self, widget=None, event=None):
        """ Change current screenshot

        Parameters
        ----------
        widget : Gtk.Widget, optional
            Object which receive signal (Default: None)
        event : Gdk.EventButton or Gdk.EventKey, optional
            Event which triggered this signal (Default: None)
        """

        # Keyboard
        if widget is self.window:
            if event.keyval == Gdk.KEY_Left:
                self.index -= 1

            elif event.keyval == Gdk.KEY_Right:
                self.index += 1

            elif event.keyval == Gdk.KEY_KP_Subtract:
                self.zoom_fit = False
                self.zoom_actual -= self.zoom_step

            elif event.keyval == Gdk.KEY_KP_Add:
                self.zoom_fit = False
                self.zoom_actual += self.zoom_step

        # Zoom
        elif widget == self.button_zoom_minus:
            self.zoom_fit = False
            self.zoom_actual -= self.zoom_step

        elif widget == self.button_original:
            self.zoom_fit = False
            self.zoom_actual = None

        elif widget == self.button_fit:
            self.zoom_fit = True

        elif widget == self.button_zoom_plus:
            self.zoom_fit = False
            self.zoom_actual += self.zoom_step

        # Move
        elif widget == self.button_first:
            self.index = 0

        elif widget in (self.button_overlay_previous, self.button_previous):
            self.index -= 1

        elif widget in (self.button_overlay_next, self.button_next):
            self.index += 1

        elif widget == self.button_last:
            self.index = len(self.screenshots) - 1

        # Fixes
        if self.index < 0:
            self.index = 0

        elif self.index > len(self.screenshots) - 1:
            self.index = len(self.screenshots) - 1

        if self.zoom_actual is not None and type(self.zoom_actual) is int:
            if self.zoom_actual < self.zoom_min:
                self.zoom_actual = self.zoom_min

            elif self.zoom_actual > self.zoom_max:
                self.zoom_actual = self.zoom_max

        self.update_screenshot()
        self.set_widgets_sensitive()

    def set_widgets_sensitive(self):
        """ Refresh interface's widgets
        """

        self.button_first.set_sensitive(True)
        self.button_previous.set_sensitive(True)
        self.button_next.set_sensitive(True)
        self.button_last.set_sensitive(True)

        if self.index == 0:
            self.button_first.set_sensitive(False)
            self.button_previous.set_sensitive(False)

            self.button_overlay_previous.hide()

        else:
            self.button_overlay_previous.show()

        if self.index == len(self.screenshots) - 1:
            self.button_last.set_sensitive(False)
            self.button_next.set_sensitive(False)

            self.button_overlay_next.hide()

        else:
            self.button_overlay_next.show()

    def update_adjustment(self, widget, scroll, value):
        """ Change current screenshot size

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        scroll : Gtk.ScrollType
            Type of scroll action that was performed
        value : float
            New value resulting from the scroll action
        """

        if int(value) >= self.zoom_min and int(value) <= self.zoom_max:
            self.zoom_fit = False
            self.zoom_actual = int(value)

            self.update_screenshot()

    def update_screenshot(self, widget=None, event=None):
        """ Change current screenshot size

        Parameters
        ----------
        widget : Gtk.Widget, optional
            Object which receive signal (Default: None)
        event : Gdk.EventButton or Gdk.EventKey, optional
            Event which triggered this signal (Default: None)
        """

        # Get the current screenshot path
        path = self.screenshots[self.index].expanduser()

        # Avoid to recreate a pixbuf for the same file
        if not str(path) == str(self.current_path):

            if path.exists() and path.is_file():
                self.current_path = path

                try:
                    # Generate a Pixbuf from current screenshot
                    self.current_pixbuf = GdkPixbuf.Pixbuf.new_from_file(
                        str(self.current_path))

                    self.current_pixbuf_size = (
                        self.current_pixbuf.get_width(),
                        self.current_pixbuf.get_height())

                    self.current_pixbuf_zoom = None

                    # Set headerbar subtitle with current screenshot path
                    self.set_subtitle(str(self.current_path))

                except GLib.Error:
                    self.current_pixbuf = None

        # Check if pixbuf has been generate correctly
        if self.current_pixbuf is not None:

            # Restore original image size
            width, height = self.current_pixbuf_size

            if width > height:
                screen_height = int((height * self.__default_size) / width)
                screen_width = self.__default_size

            else:
                screen_width = int((width * self.__default_size) / height)
                screen_height = self.__default_size

            # ------------------------------------
            #   Check zoom features
            # ------------------------------------

            # Zoom to original size
            if self.zoom_actual is None:
                ratio_x = float(width / screen_width)
                ratio_y = float(height / screen_height)

                self.zoom_actual = int(min(ratio_x, ratio_y) * 100)

            # Zoom to fit current window
            elif self.zoom_fit:
                allocation = self.scroll_image.get_allocation()

                ratio_x = float(allocation.width / screen_width)
                ratio_y = float(allocation.height / screen_height)

                self.zoom_actual = int(min(ratio_x, ratio_y) * 100)

            # ------------------------------------
            #   Reload pixbuf
            # ------------------------------------

            # Avoid to have a zoom above maximum allowed
            if self.zoom_actual > self.zoom_max:
                self.zoom_actual = self.zoom_max
            # Avoid to have a zoom under minimum allowed
            if self.zoom_actual < self.zoom_min:
                self.zoom_actual = self.zoom_min

            if not self.current_pixbuf_zoom == self.zoom_actual:
                self.current_pixbuf_zoom = self.zoom_actual

                # Calc ratio from current zoom percent
                ratio_width = int((self.zoom_actual * screen_width) / 100)
                ratio_height = int((self.zoom_actual * screen_height) / 100)

                # Update scale adjustment
                self.adjustment_zoom.set_value(float(self.zoom_actual))

                self.image.set_from_pixbuf(self.current_pixbuf.scale_simple(
                    ratio_width, ratio_height, GdkPixbuf.InterpType.TILES))



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Datetime
from datetime import datetime, timedelta

# Filesystem
from os import access, R_OK
from os.path import getctime
from pathlib import Path
from shutil import copy2

# Regex
from re import sub

# System
from os import environ
from sys import version_info


# ------------------------------------------------------------------------------
#   Methods
# ------------------------------------------------------------------------------

def get_data(*args, egg="geode_gem"):
    """ Provides easy access to data in a python egg or local folder

    This function search a path in a specific python egg or in local folder.
    The local folder is check before egg to allow quick debugging.

    Thanks Deluge :)

    Parameters
    ----------
    args : list
        File path as strings list
    egg : str, optional
        Python egg name (Default: 'geode_gem')

    Returns
    -------
    pathlib.Path
        File path instance
    """

    path = Path(*args).expanduser()

    try:
        data = resource_from_path(egg, *args)

    except ImportError:
        from pkg_resources import resource_filename

        data = Path(resource_filename(egg, str(path))).expanduser()

    except Exception:
        data = Path(egg).joinpath(path)

    if data.exists():
        return data

    return path


def resource_from_path(egg, *args):
    """ Retrieve path from Python module resource data

    Notes
    -----
    Only available for Python >= 3.7

    Parameters
    ----------
    egg : str
        Python egg name
    args : list
        File path as strings list

    Returns
    -------
    pathlib.Path
        Resource path object
    """

    from importlib import resources

    # Generate a module name based on egg name and file resource path
    module_name = '.'.join([egg, *args[:-1]])

    # Try to retrieve Path from module resource
    try:
        with resources.path(module_name, args[-1]) as filepath:
            return filepath.expanduser()

    # Then retrieve Path from anywhere into the main module
    except IsADirectoryError:
        with resources.path(egg, "__init__.py") as filepath:
            return filepath.parent.joinpath(*args).expanduser()


def parse_timedelta(delta):
    """ Parse a deltatime to string

    Get a string from the deltatime formated as HH:MM:SS

    Parameters
    ----------
    delta : datetime.timedelta
        Deltatime to parse

    Returns
    -------
    str or None
        Parse value
    """

    if delta is None:
        return None

    hours, minutes, seconds = int(), int(), int()

    if type(delta) is timedelta:
        hours, seconds = divmod(delta.seconds, 3600)

        if seconds > 0:
            minutes, seconds = divmod(seconds, 60)

        hours += delta.days * 24

    return "{:02d}:{:02d}:{:02d}".format(hours, minutes, seconds)


def get_binary_path(binary):
    """ Get a list of available binary paths from $PATH variable

    This function get all the path from $PATH variable which match binary
    request.

    Parameters
    ----------
    binary : str
        Binary name or path

    Returns
    -------
    list
        List of available path

    Examples
    --------
    >>> get_binary_path("ls")
    ['/bin/ls']
    """

    available = list()

    if binary is None or len(str(binary)) == 0:
        return available

    binary = Path(binary).expanduser()

    if binary.exists():
        available.append(binary.name)

    for directory in set(environ["PATH"].split(':')):
        path = Path(directory).expanduser()

        if access(path, R_OK):
            binary_path = path.joinpath(binary)

            if binary_path.exists() and binary_path.name not in available:
                available.append(str(binary_path))

    return available


def generate_identifier(path):
    """ Generate an identifier from a path

    Parameters
    ----------
    path : pathlib.Path or str
        Path to parse into indentifier

    Returns
    -------
    str
        Identifier string

    Examples
    --------
    >>> generate_identifier("Double Dragon (Europe).nes")
    'double-dragon-europe-nes-25953832'
    """

    inode = int()

    if isinstance(path, str):
        path = Path(path).expanduser()

    if path.exists():
        # Retrieve file inode number
        inode = path.stat().st_ino

    # Retrieve file basename
    path = path.name

    # Retrieve only alphanumeric element from filename
    name = sub(r"[^\w\d]+", ' ', path.lower())
    # Remove useless spaces and replace the others with a dash
    name = sub(r"[\s|_]+", '-', name.strip())

    if inode > 0:
        name = f"{name}-{inode}"

    return name


def generate_extension(extension):
    """ Generate a regex pattern to check lower and upper case extensions

    Thanks to https://stackoverflow.com/a/10148272

    Parameters
    ----------
    extension : str
        Extension to parse without the first dot

    Returns
    -------
    str
        Regex pattern

    Examples
    --------
    >>> generate_extension("nes")
    '[nN][eE][sS]'
    """

    pattern = str()

    for character in extension:
        if not character == '.':
            pattern += f"[{character.lower()}{character.upper()}]"

        else:
            pattern += '.'

    return pattern


def copy(src, dst, follow_symlinks=True):
    """ Compat function to use shutil.copy on python < 3.6

    More informations with shutil.copy2 module

    Parameters
    ----------
    src : str or pathlib.Path
        File source
    dst : str or pathlib.Path
        File destination
    follow_symlinks : bool, optional
        Retrieve metadata from symlinks origins if True
    """

    if version_info.major == 3 and version_info.minor < 6:

        if not isinstance(src, str):
            src = str(src)

        if not isinstance(dst, str):
            dst = str(dst)

    copy2(src, dst, follow_symlinks=follow_symlinks)


def get_creation_datetime(path):
    """ Retrieve the creation date from a specific filename

    Parameters
    ----------
    path : pathlib.Path or str
        Path to retrieve creation datetime

    Returns
    -------
    datetime.datetime
        Creation datetime object

    Examples
    --------
    >>> get_creation_datetime("~/.bashrc")
    datetime.datetime(2019, 9, 22, 14, 1, 37, 56527)
    """

    if isinstance(path, str):
        path = Path(path).expanduser()

    if not path.exists():
        return None

    return datetime.fromtimestamp(getctime(path))


def get_boot_datetime_as_timestamp(proc_path="/proc"):
    """ Retrieve boot datetime as timestamp

    The 'uptime' file contains two values: boot time and idle time. This method
    only retrieve the first one to calculate the boot datetime.

    Parameters
    ----------
    proc_path : pathlib.Path or str, optional
        Process file system path (Only used to test the method)

    Returns
    -------
    float
        Boot datetime as timestamp value

    Raises
    ------
    FileNotFoundError
        When the /proc directory do not exists on filesystem
        When the /proc/uptime file do not exists on filesystem
    """

    if isinstance(proc_path, str):
        proc_path = Path(proc_path)

    if not proc_path.exists():
        raise FileNotFoundError("Cannot found process file system")

    uptime_file = proc_path.joinpath("uptime")
    if not uptime_file.exists():
        raise FileNotFoundError(f"Cannot found {uptime_file} on filesystem")

    with uptime_file.open('r') as pipe:
        content = pipe.read().split('\n')[0]

    if content:
        return datetime.now().timestamp() - float(content.split()[0])

    return None


def are_equivalent_timestamps(first, second, delta=0):
    """ Check if two timestamps are equivalent

    Parameters
    ----------
    first : int
        First timestamp
    second : int
        Second timestamp
    delta : int, optional
        Allowed difference between the two timestamps

    Returns
    -------
    bool
        True if timestamps are equivalent, False otherwise
    """

    return abs(int(float(first)) - int(float(second))) in range(0, delta + 1)
